<?php

	/**
	 *  NOTES FROM NICK
	 * --------------------------------------------------
	 *   As requested, I have gone pretty heavy on the
	 *   comments and verbosity for this file.
	 *
	 *   Where applicable, I have added a `@source` for
	 *   snippets of code: the first being how the code
	 *   was inserted—via Internet searches or my collection
	 *   of snippets—followed by the original source.
	 *
	 *   When I find a function or useful bit of code on
	 *   StackOverflow, CSS-Tricks, or CodePen I will
	 *   typically turn it into a code snippet in Atom
	 *   to ensure reusability.
	 *
	 *   Less importantly: I acknowledge that I am
	 *   "cheating" a bit by formatting all of the
	 *   data on the server-side. If the data structure
	 *   were to be "untrustworthy", I would end the
	 *   script after the `get_data` method and use
	 *   JavaScript to handle the data processing. This
	 *   would take longer to process, but a loading
	 *   screen would be better-receieved than a
	 *   blank, loading browser while the back-end
	 *   generates and processes the data.
	 * --------------------------------------------------
	 */

	class MLBDataMiner {

		function __construct() {

			global $source_url;

			$source_url = "https://questionnaire-148920.appspot.com/swe/data.html";

			/** Set up CORS settings to circumvent cross-origin error. */
			$this->start_cors();

			/** Get data. */
			$this->data = $this->get_data($source_url);

			/** Format data. */
			$this->formatted = $this->format_data($this->data);

			/** Sort data by salary, from highest to lowest. */
			/** @source snippet */
			/** @source https://stackoverflow.com/questions/2699086/how-to-sort-multi-dimensional-array-by-value */
			usort($this->formatted['data'], function($a, $b) {
    		return $a[1] < $b[1];
			});

			/** Calculate qualifying offer. */
			$this->qualifying_offer_data = $this->calculate_qualifying_offer($this->formatted);

			/** Echo JSON to page. */
			echo json_encode($this->qualifying_offer_data);

		}



		/**
		 *  Calculates qualifying offer.
		 *  @param  [array] $arr  Salary array.
		 *  @return [array]       Salary array with qualifying offer.
		 */
		private function calculate_qualifying_offer($arr) {
			$total_top_salary = 0;

			for ($i = 0; $i <= 124; $i++) {
				$s_arr[] = $arr['data'][$i][1];
				$total_top_salary = $total_top_salary + $arr['data'][$i][1];
			}
			$arr['salary']['qualifying_offer'] = ceil($total_top_salary / 125);
			return $arr;
		}



		/**
		 *  Filters raw HTML into a usable array.
		 *  @param  [object] $html Raw HTML from cURL pull.
		 *  @return [array]        Array containing salary data.
		 */
		private function filter_html($html) {

			$data = [];

			/** Use DOMDocument and DOMXpath to filter the raw HTML to
				* just the salary table.
				* @source Internet search
	 		  * @source https://www.php.net/manual/en/class.domdocument.php
			  */
			$dom = new DOMDocument();
			$dom->loadHTML($html);
			$xpath = new DOMXpath($dom);
			$salary_table_rows = $xpath->query('//table[@id="salaries-table"]//tbody//tr');

			/** Convert table rows and cells to arrays. */
			foreach($salary_table_rows as $row => $tr) {
				foreach($tr->childNodes as $td) {
					$data[$row][] = preg_replace('~[\r\n]+~', '', trim($td->nodeValue));
				}
				$data[$row] = array_values(array_filter($data[$row]));
			}

			/** Return the salary table data, converted into an array. */
			return $data;
		}



		/**
		 *  Format usable data into array.
		 *  @param  [array] $salary_arr Unformatted array of salary data.
		 *  @return [array]             Formatted (but incomplete) array of salary data.
		 */
		private function format_data($salary_arr) {

			/** Create the top-level sections for final JSON output. */
			$data = [];
			$data['data'] = [];
			$data['salary'] = [];

			/** Set up counter for NULL salary entries.  */
			$no_salary_count = 0;

			/** Total amount paid out to players */
			$total_salary = 0;

			/** Loop through each row in the salary array. */
			foreach($salary_arr as $row) {

				$data_row = [];
				$arr_length = sizeof($row);

				/** If the array has three items, it's missing salary data. */
				if ($arr_length == 3) {
					/** Increase the NULL salary counter. */
					$no_salary_count = $no_salary_count + 1;
					$data_row = [
						$row[0],
						null,
						$row[1],
						$row[2]
					];
				}

				/** If the array has four items, it's complete. */
				if ($arr_length == 4) {
					$formatted_salary = $this->format_salary_data($row[1]);
					$total_salary = $total_salary + $formatted_salary;
					$data_row = [
						$row[0],
						$formatted_salary,
						$row[2],
						$row[3]
					];
				}

				$data['data'][] = $data_row;
			}

			/** Add count for missing salary data. */
			$data['salary']['missing_data_count'] = $no_salary_count;

			/** Add league average salary */
			$data['salary']['league_average'] = ceil($total_salary / (sizeof($data['data']) - $no_salary_count));

			/** Return array. */
			return $data;
		}



		/**
		 *  Removes any non-numeric character from a string.
		 *  @source snippet (modified)
		 *  @source https://regexr.com/
		 */
		private function format_salary_data($salary_string) {
			return (int)preg_replace('/\D/', '', $salary_string);
		}



		/**
		 *  Retrieves data from web page and converts into usable data.
		 *  @param  [string] $url URL to web page with salary info.
		 *  @return [array]       Array with unformatted salary data.
		 */
		private function get_data($url) {
			/** Get raw HTML via cURL */
			$html = $this->get_html($url);
			/** Filter elements using DOMDocument and DOMXpath */
			$salary_arr = $this->filter_html($html);
			/** Return dirty array. */
			return $salary_arr;
		}



		/**
		 *  Pulls HTML from URL via cURL.
		 *  @source snippet
		 *  @source https://www.phpzag.com/how-to-scrape-data-from-website-using-php
		 */
		private function get_html($url) {
			$ch = curl_init($url);
			curl_setopt($ch, CURLOPT_RETURNTRANSFER, true);
			return curl_exec($ch);
		}



		/**
		 *  Executes CORS settings to circumvent cross-origin error.
		 *  @source snippet
		 *  @source https://stackoverflow.com/questions/8719276/cross-origin-request-headerscors-with-php-headers
		 */
		private function start_cors() {
			/** Allow from any origin */
			if (isset($_SERVER['HTTP_ORIGIN'])) {
				header("Access-Control-Allow-Origin: {$_SERVER['HTTP_ORIGIN']}");
				header('Access-Control-Allow-Credentials: true');
				header('Access-Control-Max-Age: 86400');
			}

			/** Access-Control headers are received during OPTIONS requests */
			if ($_SERVER['REQUEST_METHOD'] == 'OPTIONS') {
				if (isset($_SERVER['HTTP_ACCESS_CONTROL_REQUEST_METHOD'])) {
					header("Access-Control-Allow-Methods: GET, POST, OPTIONS");
				}

				if (isset($_SERVER['HTTP_ACCESS_CONTROL_REQUEST_HEADERS'])) {
					header("Access-Control-Allow-Headers: {$_SERVER['HTTP_ACCESS_CONTROL_REQUEST_HEADERS']}");
				}

				exit(0);
			}
		}

	/** End of class. */
	}


/** --- Execute mine. ------ */
new MLBDataMiner();
